from setuptools import setup

import elifetools

with open("README.rst") as fp:
    README = fp.read()

setup(
    name="elifetools",
    version=elifetools.__version__,
    description="Tools for using article data in Python.",
    long_description=README,
    long_description_content_type="text/x-rst",
    packages=["elifetools"],
    license="MIT",
    install_requires=[
        "beautifulsoup4",
        "lxml",
        "python-slugify",
    ],
    url="https://github.com/elifesciences/elife-tools",
    maintainer="eLife Sciences Publications Ltd.",
    maintainer_email="tech-team@elifesciences.org",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
)



from elifetools.utils import (
    first,
    firstnn,
    lazy_extract_nodes,
    extract_first_node,
    extract_nodes,
    node_contents_str,
)

"""
rawParser.py extracts and returns the nodes from the article xml using BeautifulSoup so that functionality at higher levels may use and combine them as neccessary.

"""


def article_meta(soup):
    return extract_first_node(soup, "article-meta")


def article_title(soup):
    return extract_first_node(soup, "article-title")


def title(soup):
    return extract_first_node(soup, "title")


def lazy_abstract(soup, abstract_type=None):
    if abstract_type:
        return lazy_extract_nodes(
            soup, "abstract", attr="abstract-type", value=abstract_type
        )
    else:
        return lazy_extract_nodes(soup, "abstract")


def abstract(soup, abstract_type=None):
    return list(lazy_abstract(soup, abstract_type))


def lazy_article_id(soup, pub_id_type=None):
    if pub_id_type:
        return lazy_extract_nodes(soup, "article-id", attr="pub-id-type", value=pub_id_type)
    else:
        return lazy_extract_nodes(soup, "article-id")


def article_id(soup, pub_id_type=None):
    return list(lazy_article_id(soup, pub_id_type))


def meta_article_id(soup, pub_id_type, specific_use=None):
    for tag in lazy_article_id(soup, pub_id_type):
        # filter by specific-use attribute
        if tag.get("specific-use") == specific_use:
            # the first article-id tag whose parent is article-meta
            if tag.parent.name == "article-meta":
                return tag
    return None


def doi(soup):
    return meta_article_id(soup, pub_id_type="doi")


def version_doi(soup):
    return meta_article_id(soup, pub_id_type="doi", specific_use="version")


def publisher_id(soup):
    return meta_article_id(soup, pub_id_type="publisher-id")


def journal_id(soup):
    # the first non-nil tag
    return firstnn(
        lazy_extract_nodes(soup, "journal-id", attr="journal-id-type", value="publisher-id")
    )


def journal_title(soup):
    return extract_first_node(soup, "journal-title")


def journal_issn(soup, pub_format, pub_type):
    if pub_format is None and pub_type is None:
        # return the first issn tag found regardless of which type
        return extract_first_node(soup, "issn")
    elif pub_format is not None:
        return extract_first_node(
            soup, "issn", attr="publication-format", value=pub_format
        )

    elif pub_type is not None:
        return extract_first_node(soup, "issn", attr="pub-type", value=pub_type)


def publisher(soup):
    return extract_first_node(soup, "publisher-name")


def article_type(soup):
    # returns raw data, just that the data doesn't contain any BS nodes
    return extract_first_node(soup, "article").get("article-type")


def article_version(soup):
    return lazy_extract_nodes(soup, "article-version")


def lazy_pub_date(soup, date_type=None, pub_type=None):
    if date_type is not None:
        return lazy_extract_nodes(soup, "pub-date", attr="date-type", value=date_type)
    elif pub_type is not None:
        return lazy_extract_nodes(soup, "pub-date", attr="pub-type", value=pub_type)
    else:
        return lazy_extract_nodes(soup, "pub-date")


def pub_date(soup, date_type=None, pub_type=None):
    return list(lazy_pub_date(soup, date_type, pub_type))


def lazy_date(soup, date_type=None):
    if date_type is not None:
        return lazy_extract_nodes(soup, "date", attr="date-type", value=date_type)
    else:
        return lazy_extract_nodes(soup, "date")


def date(soup, date_type=None):
    return list(lazy_date(soup, date_type))


def history_date(soup, date_type):
    for date_tag in lazy_date(soup, date_type):
        if date_tag.parent.name == "history":
            return date_tag


def day(soup):
    return extract_first_node(soup, "day")


def month(soup):
    return extract_first_node(soup, "month")


def year(soup):
    return extract_first_node(soup, "year")


def lazy_keyword_group(soup):
    return lazy_extract_nodes(soup, "kwd-group")


def keyword_group(soup):
    return list(lazy_keyword_group(soup))


def acknowledgements(soup):
    return extract_first_node(soup, "ack")


def lazy_conflict(soup):
    for tag in lazy_extract_nodes(soup, "fn", attr="fn-type", value="conflict"):
        yield tag
    for tag in lazy_extract_nodes(soup, "fn", attr="fn-type", value="COI-statement"):
        yield tag


def conflict(soup):
    return list(lazy_conflict(soup))


def lazy_permissions(soup):
    # a better selector might be "article-meta.permissions"
    return lazy_extract_nodes(soup, "permissions")


def permissions(soup):
    return list(lazy_permissions(soup))


def article_permissions(soup):
    # a better selector might be "article-meta.permissions"
    for tag in lazy_permissions(soup):
        if tag.parent.name == "article-meta":
            return tag


def lazy_licence(soup):
    return lazy_extract_nodes(soup, "license")


def licence(soup):
    return list(lazy_licence(soup))


def lazy_licence_p(soup):
    return lazy_extract_nodes(soup, "license-p")


def licence_p(soup):
    return list(lazy_licence_p(soup))


def licence_url(soup):
    "License url attribute of the license tag"
    first_licence = first(lazy_licence(soup))
    if first_licence:
        return first_licence.get("xlink:href")


def attrib(soup):
    return extract_nodes(soup, "attrib")


def copyright_statement(soup):
    return extract_first_node(soup, "copyright-statement")


def copyright_year(soup):
    return extract_first_node(soup, "copyright-year")


def copyright_holder(soup):
    return extract_first_node(soup, "copyright-holder")


def funding_statement(soup):
    return extract_first_node(soup, "funding-statement")


def affiliation(soup):
    return extract_nodes(soup, "aff")


def research_organism_keywords(soup):
    tags = extract_first_node(
        soup, "kwd-group", attr="kwd-group-type", value="research-organism"
    )

    if not tags:
        return None
    return [tag for tag in tags if tag.name == "kwd"] or None


def author_keywords(soup):
    # A few articles have kwd-group with no kwd-group-type, so account for those
    tags = extract_nodes(soup, "kwd-group")
    keyword_tags = []
    for tag in tags:
        if (
            tag.get("kwd-group-type") == "author-keywords"
            or tag.get("kwd-group-type") is None
        ):
            keyword_tags += [tag for tag in tag if tag.name == "kwd"]
    return keyword_tags


def lazy_subject_area(soup, subject_group_type=None):
    # Supports all subject areas or just particular ones filtered by
    tags = lazy_extract_nodes(soup, "subject")

    if subject_group_type:
        for tag in tags:
            if tag.parent.get("subj-group-type") == subject_group_type:
                yield tag
    else:
        for tag in tags:
            if (
                tag.parent.name == "subj-group" and
                tag.parent.parent.name == "article-categories" and
                tag.parent.parent.parent.name == "article-meta"
            ):
                yield tag


def subject_area(soup, subject_group_type=None):
    return list(lazy_subject_area(soup, subject_group_type))


def full_subject_area(soup, subject_group_type=None):

    subject_group_tags = extract_nodes(soup, "subj-group")
    subject_group_tags = [
        tag
        for tag in subject_group_tags
        if tag.parent.name == "article-categories"
        and tag.parent.parent.name == "article-meta"
    ]

    if subject_group_type:
        subject_group_tags = list(
            filter(lambda tag: tag.get("subj-group-type" == subject_group_type))
        )

    return subject_group_tags


def lazy_custom_meta(soup, meta_name=None):
    custom_meta_tags = lazy_extract_nodes(soup, "custom-meta")
    if meta_name is not None:
        for tag in custom_meta_tags:
            if node_contents_str(extract_first_node(tag, "meta-name")) == meta_name:
                yield tag
    else:
        for tag in custom_meta_tags:
            yield tag


def custom_meta(soup, meta_name=None):
    return list(lazy_custom_meta(soup, meta_name))


def display_channel(soup):
    return subject_area(soup, subject_group_type="display-channel")


def lazy_sub_display_channel(soup):
    return lazy_subject_area(soup, subject_group_type="sub-display-channel")


def sub_display_channel(soup):
    return list(lazy_sub_display_channel(soup))


def category(soup):
    return subject_area(soup, subject_group_type="heading")


def related_article(soup):
    related_article_tags = extract_nodes(soup, "related-article")
    return [tag for tag in related_article_tags if tag.parent.name == "article-meta"]


def lazy_mixed_citations(soup):
    return lazy_extract_nodes(soup, "mixed-citation")


def mixed_citations(soup):
    return list(lazy_mixed_citations(soup))


def related_object(soup):
    return extract_nodes(soup, "related-object")


def lazy_object_id(soup, pub_id_type):
    return lazy_extract_nodes(soup, "object-id", attr="pub-id-type", value=pub_id_type)


def object_id(soup, pub_id_type):
    return list(lazy_object_id(soup, pub_id_type))


def lazy_pub_history(soup):
    return lazy_extract_nodes(soup, "pub-history")


def pub_history(soup):
    return list(lazy_pub_history(soup))


def event(soup):
    return extract_nodes(soup, "event")


def lazy_event_desc(soup):
    return lazy_extract_nodes(soup, "event-desc")


def event_desc(soup):
    return list(lazy_event_desc(soup))


def label(soup):
    return extract_first_node(soup, "label")


def contributors(soup):
    return extract_nodes(soup, "contrib")


def article_contributors(soup):
    article_meta_tag = article_meta(soup)
    if article_meta_tag:
        contributor_tags = extract_nodes(article_meta_tag, ["contrib", "on-behalf-of"])
        return [tag for tag in contributor_tags if tag.parent.name == "contrib-group"]


def authors(soup, contrib_type="author"):
    if contrib_type:
        return extract_nodes(soup, "contrib", attr="contrib-type", value=contrib_type)
    else:
        return extract_nodes(soup, "contrib")


def caption(soup):
    return extract_first_node(soup, "caption")


def author_notes(soup):
    return extract_first_node(soup, "author-notes")


def corresp(soup):
    return extract_nodes(soup, "corresp")


def lazy_fn_group(soup, content_type=None):
    if content_type:
        return lazy_extract_nodes(soup, "fn-group", attr="content-type", value=content_type)
    else:
        return lazy_extract_nodes(soup, "fn-group")


def fn_group(soup, content_type=None):
    return list(lazy_fn_group(soup, content_type))


def fn(soup):
    return extract_nodes(soup, "fn")


def lazy_media(soup):
    return lazy_extract_nodes(soup, "media")


def media(soup):
    return list(lazy_media(soup))


def lazy_inline_graphic(soup):
    return lazy_extract_nodes(soup, "inline-graphic")


def inline_graphic(soup):
    return list(lazy_inline_graphic(soup))


def lazy_graphic(soup):
    return lazy_extract_nodes(soup, "graphic")


def graphic(soup):
    return list(lazy_graphic(soup))


def lazy_self_uri(soup):
    return lazy_extract_nodes(soup, "self-uri")


def self_uri(soup):
    return list(lazy_self_uri(soup))


def lazy_supplementary_material(soup):
    return lazy_extract_nodes(soup, "supplementary-material")


def supplementary_material(soup):
    return list(lazy_supplementary_material(soup))


#
# authors
#

def lazy_contrib_id(soup):
    return lazy_extract_nodes(soup, "contrib-id")


def contrib_id(soup):
    return list(lazy_contrib_id(soup))


def email(soup):
    return extract_nodes(soup, "email")


def lazy_phone(soup):
    return lazy_extract_nodes(soup, "phone")


def phone(soup):
    return list(lazy_phone(soup))


def lazy_bio(soup):
    return lazy_extract_nodes(soup, "bio")


def bio(soup):
    return list(lazy_bio(soup))


#
# references
#


def ref_list(soup):
    return extract_nodes(soup, "ref")


def lazy_volume(soup):
    return lazy_extract_nodes(soup, "volume")


def volume(soup):
    return list(lazy_volume(soup))


def lazy_issue(soup):
    return lazy_extract_nodes(soup, "issue")


def issue(soup):
    return list(lazy_issue(soup))


def lazy_elocation_id(soup):
    return lazy_extract_nodes(soup, "elocation-id")


def elocation_id(soup):
    return list(lazy_elocation_id(soup))


def lazy_fpage(soup):
    return lazy_extract_nodes(soup, "fpage")    


def fpage(soup):
    return list(lazy_fpage(soup))


def lazy_lpage(soup):
    return lazy_extract_nodes(soup, "lpage")


def lpage(soup):
    return list(lazy_lpage(soup))


def lazy_collab(soup):
    return lazy_extract_nodes(soup, "collab")


def collab(soup):
    return list(lazy_collab(soup))


def lazy_publisher_loc(soup):
    return lazy_extract_nodes(soup, "publisher-loc")


def publisher_loc(soup):
    return list(lazy_publisher_loc(soup))


def lazy_publisher_name(soup):
    return lazy_extract_nodes(soup, "publisher-name")


def publisher_name(soup):
    return list(lazy_publisher_name(soup))


def lazy_comment(soup):
    return lazy_extract_nodes(soup, "comment")


def comment(soup):
    return list(lazy_comment(soup))


def lazy_element_citation(soup):
    return lazy_extract_nodes(soup, "element-citation")


def element_citation(soup):
    return list(lazy_element_citation(soup))


def lazy_etal(soup):
    return lazy_extract_nodes(soup, "etal")


def etal(soup):
    return list(lazy_etal(soup))


def lazy_pub_id(soup, pub_id_type=None):
    if pub_id_type:
        return lazy_extract_nodes(soup, "pub-id", attr="pub-id-type", value=pub_id_type)
    else:
        return lazy_extract_nodes(soup, "pub-id")


def pub_id(soup, pub_id_type=None):
    return list(lazy_pub_id(soup, pub_id_type))


def lazy_source(soup):
    return lazy_extract_nodes(soup, "source")


def source(soup):
    return list(lazy_source(soup))


def person_group(soup):
    return extract_nodes(soup, "person-group")


def lazy_surname(soup):
    return lazy_extract_nodes(soup, "surname")


def surname(soup):
    return list(lazy_surname(soup))


def lazy_given_names(soup):
    return lazy_extract_nodes(soup, "given-names")


def given_names(soup):
    return list(lazy_given_names(soup))


def lazy_suffix(soup):
    return lazy_extract_nodes(soup, "suffix")


def suffix(soup):
    return list(lazy_suffix(soup))


def lazy_ext_link(soup, ext_link_type=None):
    if ext_link_type:
        return lazy_extract_nodes(
            soup, "ext-link", attr="ext-link-type", value=ext_link_type
        )
    else:
        return lazy_extract_nodes(soup, "ext-link")


def ext_link(soup, ext_link_type=None):
    return list(lazy_ext_link(soup, ext_link_type))


def lazy_uri(soup):
    return lazy_extract_nodes(soup, "uri")


def uri(soup):
    return list(lazy_uri(soup))


def lazy_edition(soup):
    return lazy_extract_nodes(soup, "edition")


def edition(soup):
    return list(lazy_edition(soup))


def lazy_version(soup):
    return lazy_extract_nodes(soup, "version")


def version(soup):
    return list(lazy_version(soup))

def lazy_chapter_title(soup):
    return lazy_extract_nodes(soup, "chapter-title")


def chapter_title(soup):
    return list(lazy_chapter_title(soup))


def lazy_data_title(soup):
    return lazy_extract_nodes(soup, "data-title")


def data_title(soup):
    return list(lazy_data_title(soup))


def lazy_conf_name(soup):
    return lazy_extract_nodes(soup, "conf-name")


def conf_name(soup):
    return list(lazy_conf_name(soup))


def lazy_date_in_citation(soup):
    return lazy_extract_nodes(soup, "date-in-citation")


def date_in_citation(soup):
    return list(lazy_date_in_citation(soup))


def lazy_patent(soup):
    return lazy_extract_nodes(soup, "patent")


def patent(soup):
    return list(lazy_patent(soup))


#
# back
#


def back(soup):
    return extract_first_node(soup, "back")


def lazy_app_group(soup):
    return lazy_extract_nodes(soup, "app-group")


def app_group(soup):
    return list(lazy_app_group(soup))


def app(soup):
    return extract_nodes(soup, "app")


#
# body
#


def body(soup):
    return extract_nodes(soup, "body")


def article_body(soup):
    return extract_first_node(soup, "body")


#
# sub-article
#


def sub_article(soup, article_type=None):
    return extract_nodes(soup, "sub-article", attr="article-type", value=article_type)


def filter_sub_articles_by_title(
    sub_article_nodes, include_title=None, exclude_title=None
):
    "filter a list of sub-article nodes by case-insensitive article-title matching"
    node_list = []

    if not sub_article_nodes:
        return node_list

    for node in sub_article_nodes:

        # only filter by the title value if there is one
        if article_title(node):
            if include_title and exclude_title:
                # test if include_title matches and exclude_title does not match
                if (
                    include_title.lower()
                    in node_contents_str(article_title(node)).lower()
                ) and (
                    exclude_title.lower()
                    not in node_contents_str(article_title(node)).lower()
                ):
                    node_list.append(node)
            elif include_title:
                # if only include_title matches
                if (
                    include_title.lower()
                    in node_contents_str(article_title(node)).lower()
                ):
                    node_list.append(node)
            elif exclude_title:
                # if only exclude_title does not match
                if (
                    exclude_title.lower()
                    not in node_contents_str(article_title(node)).lower()
                ):
                    node_list.append(node)
            else:
                # by default keep the node
                node_list.append(node)
        else:
            # by default keep the node if it has no title regardless of filter term
            node_list.append(node)

    return node_list


def editor_report_by_title(soup, include_title=None):
    "list of editor-report sub-article filtered by article-title value"
    sub_article_nodes = sub_article(soup, "editor-report")
    return filter_sub_articles_by_title(sub_article_nodes, include_title=include_title)


def editor_evaluation(soup):
    "sub-article node of type editor-report with evaluation in the article-title"
    return first(editor_report_by_title(soup, include_title="evaluation"))


def elife_assessment(soup):
    "sub-article node of type editor-report with assessment in the article-title"
    return first(editor_report_by_title(soup, include_title="assessment"))


def decision_letter(soup):
    tag = first(sub_article(soup, "article-commentary"))
    if not tag:
        tag = first(sub_article(soup, "decision-letter"))
    return tag


def author_response(soup):
    "sub-article node of type reply or author-comment, return the first one found"
    return first(sub_article(soup, "reply") + sub_article(soup, "author-comment"))


def referee_report_by_title(soup, include_title=None, exclude_title=None):
    "list of referee-report sub-article filtered by article-title value"
    sub_article_nodes = sub_article(soup, "referee-report")
    return filter_sub_articles_by_title(
        sub_article_nodes, include_title=include_title, exclude_title=exclude_title
    )


def recommendations_for_authors(soup):
    "sub-article node of type referee-report with recommendations in the article-title"
    return first(referee_report_by_title(soup, include_title="recommendations"))


def public_reviews(soup):
    "list of sub-article node of type referee-report without recommendations in the article-title"
    return referee_report_by_title(soup, exclude_title="recommendations")


#
# block content
#

def lazy_section(soup, sec_type=None):
    if sec_type:
        return lazy_extract_nodes(soup, "sec", attr="sec-type", value=sec_type)
    else:
        return lazy_extract_nodes(soup, "sec")


def section(soup, sec_type=None):
    return list(lazy_section(soup, sec_type))


def lazy_paragraph(soup):
    return lazy_extract_nodes(soup, "p")


def paragraph(soup):
    return list(lazy_paragraph(soup))


def table(soup):
    return extract_nodes(soup, "table")


def table_wrap_foot(soup):
    return extract_nodes(soup, "table-wrap-foot")


def disp_formula(soup):
    return extract_nodes(soup, "disp-formula")


def lazy_math(soup):
    return lazy_extract_nodes(soup, "math")


def math(soup):
    return list(lazy_math(soup))


def lazy_boxed_text(soup):
    return lazy_extract_nodes(soup, "boxed-text")


def boxed_text(soup):
    return list(lazy_boxed_text(soup))


def fig(soup):
    return extract_nodes(soup, "fig")


def fig_group(soup):
    return extract_nodes(soup, "fig-group")


# Redefining `list` could be problematic
# lsh@2023-12-01: renamed `list` to `get_list`
def get_list(soup): 
    return extract_nodes(soup, "list")


def list_item(soup):
    return extract_nodes(soup, "list-item")


#
# funding
#

def lazy_funding_group(soup):
    return lazy_extract_nodes(soup, "funding-group")


def funding_group(soup):
    return list(lazy_funding_group(soup))


def award_group(soup):
    return extract_nodes(soup, "award-group")


def principal_award_recipient(soup):
    return extract_nodes(soup, "principal-award-recipient")


def lazy_string_name(soup):
    return lazy_extract_nodes(soup, "string-name")


def string_name(soup):
    return list(lazy_string_name(soup))





from io import BytesIO
import sys
import unittest

from xml.etree import ElementTree
from xml.etree.ElementTree import Element, SubElement
from xml.dom import minidom
from ddt import ddt, data, unpack

from elifetools import xmlio
from tests.file_utils import sample_xml, read_fixture, fixture_file


def xml_expected_namespace_fix(xml_string):
    "prior to Python 3.8 the XML attributes are alphabetised, change the test fixtures"
    if sys.version_info < (3, 8):
        xml_string = xml_string.replace(
            b'<article xmlns:xlink="http://www.w3.org/1999/xlink" article-type="research-article" dtd-version="1.1d3">',
            b'<article article-type="research-article" dtd-version="1.1d3" xmlns:xlink="http://www.w3.org/1999/xlink">',
        )
    return xml_string


@ddt
class TestXmlio(unittest.TestCase):
    def setUp(self):
        pass

    @data("elife-kitchen-sink.xml")
    def test_parse(self, filename):
        root = xmlio.parse(sample_xml(filename))
        self.assertEqual(type(root), Element)

    @data(
        "elife-02833-v2.xml",
        "simple-jats-doctype-1.1.xml",
        "simple-jats-doctype-1.1d3.xml",
    )
    def test_input_output(self, filename):
        """test parsing a file while retaining the doctype"""
        with open(sample_xml(filename), "rb") as xml_file:
            xml_output_expected = xml_file.read()
        # hack for running less than Python 3.8, the XML attributes are alphabetised
        xml_output_expected = xml_expected_namespace_fix(xml_output_expected)
        root, doctype_dict = xmlio.parse(sample_xml(filename), return_doctype_dict=True)
        self.assertEqual(xmlio.output(root, None, doctype_dict), xml_output_expected)

    @data("elife-02833-v2.xml")
    def test_input_output_forcing_jats_doctype(self, filename):
        with open(sample_xml(filename), "rb") as xml_file:
            xml_output_expected = xml_file.read()
        # hack for running less than Python 3.8, the XML attributes are alphabetised
        xml_output_expected = xml_expected_namespace_fix(xml_output_expected)
        root, doctype_dict = xmlio.parse(sample_xml(filename), return_doctype_dict=True)
        self.assertEqual(xmlio.output(root, "JATS"), xml_output_expected)

    @unpack
    @data(("xmlio_input.xml", "inline-graphic", 2), ("xmlio_input.xml", "italic", None))
    def test_get_first_element_index(self, filename, tag_name, index):
        root = xmlio.parse(sample_xml(filename))
        self.assertEqual(xmlio.get_first_element_index(root, tag_name), index)

    @unpack
    @data(("<a><b/></a>", "c", "see", "b", "<a><c>see</c><b /></a>"))
    def test_add_tag_before(self, xml, tag_name, tag_text, before, expected_string):
        xml_element = ElementTree.fromstring(xml)
        element = xmlio.add_tag_before(tag_name, tag_text, xml_element, before)
        self.assertEqual(ElementTree.tostring(element).decode("utf-8"), expected_string)

    @unpack
    @data(
        (
            {"doc1.pdf": "doc-1.pdf", "img2.tif": "img-2.tif"},
            "xmlio_input.xml",
            "xmlio_output_convert_xlink.xml",
        )
    )
    def test_convert_xlink_href(
        self, name_map, xml_input_filename, xml_expected_filename
    ):
        xmlio.register_xmlns()
        root = xmlio.parse(sample_xml(xml_input_filename))
        xlink_count = xmlio.convert_xlink_href(root, name_map)
        xml_output = xmlio.output(root)
        xml_output_expected = None
        with open(sample_xml(xml_expected_filename), "rb") as xml_file:
            xml_output_expected = xml_file.read()
        # hack for running less than Python 3.8, the XML attributes are alphabetised
        xml_output_expected = xml_expected_namespace_fix(xml_output_expected)
        self.assertEqual(xml_output, xml_output_expected)
        self.assertEqual(xlink_count, 2)

    @unpack
    @data(
        (
            b"<article/>",
            "",
            "",
            None,
            '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article><article/>',
        ),
        (
            b"<article/>",
            "-//a",
            "a",
            None,
            '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article PUBLIC "-//a"  "a"><article/>',
        ),
        (
            b"<article/>",
            None,
            "b",
            "",
            '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article SYSTEM "b"><article/>',
        ),
        (
            b"<article/>",
            "c",
            "",
            "subset",
            (
                '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article PUBLIC "c"'
                '  "" [subset]><article/>'
            ),
        ),
        (
            b"<article>\n<p/>\n<?fig-width 50%?>\n<fig/>\n</article>",
            "c",
            "",
            "subset",
            (
                '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article PUBLIC "c"'
                '  "" [subset]><article>\n<p/>\n<?fig-width 50%?>\n<fig/>\n</article>'
            ),
        ),
    )
    def test_output_root(self, xml, publicId, systemId, internalSubset, xml_expected):
        encoding = "UTF-8"
        qualifiedName = "article"
        root = xmlio.parse(BytesIO(xml), insert_pis=True)
        doctype = xmlio.build_doctype(qualifiedName, publicId, systemId, internalSubset)
        xml_output = xmlio.output_root(root, doctype, encoding)
        self.assertEqual(xml_output.decode("utf-8"), xml_expected)

    def test_append_minidom_xml_to_elementtree_xml(self):
        "test coverage of an edge case"
        parent = Element("root")
        tag = SubElement(parent, "i")
        tag.text = "Text"
        tag.tail = " and tail."
        attributes = ["class", "empty"]
        # Add a first example
        xml = '<span><i>and</i> some <i id="test">XML</i>.</span>'
        reparsed = minidom.parseString(xml.encode("utf-8"))
        parent = xmlio.append_minidom_xml_to_elementtree_xml(
            parent, reparsed, child_attributes=True
        )
        # Add another example to test more lines of code
        xml = '<span class="span" empty=""> <i>more</i> text</span>'
        reparsed = minidom.parseString(xml.encode("utf-8"))
        parent = xmlio.append_minidom_xml_to_elementtree_xml(
            parent, reparsed, False, attributes
        )
        # Generate output and compare it
        rough_string = ElementTree.tostring(parent).decode("utf-8")
        self.assertEqual(
            rough_string,
            (
                '<root><i>Text</i> and tail.<span><i>and</i> some <i id="test">XML</i>.'
                '</span><span class="span" empty=""> <i>more</i> text</span></root>'
            ),
        )

    @unpack
    @data(
        # example of removing and adding the same heading tags
        (
            fixture_file("test_xmlio", "test_rewrite_subject_group_01.xml"),
            ["Cell Biology", "Plant Biology"],
            "heading",
            True,
            read_fixture("test_xmlio", "test_rewrite_subject_group_01_expected.xml"),
        ),
        # example of removing and adding the display-channel should retain the same tag order
        (
            fixture_file("test_xmlio", "test_rewrite_subject_group_02.xml"),
            ["Research Article"],
            "display-channel",
            True,
            read_fixture("test_xmlio", "test_rewrite_subject_group_02_expected.xml"),
        ),
        # example of appending a tag
        (
            fixture_file("test_xmlio", "test_rewrite_subject_group_03.xml"),
            ["New Heading"],
            "heading",
            False,
            read_fixture("test_xmlio", "test_rewrite_subject_group_03_expected.xml"),
        ),
        # example adding a new subject type will be added to the head of the list
        (
            fixture_file("test_xmlio", "test_rewrite_subject_group_04.xml"),
            ["New Heading"],
            "sub-display-channel",
            True,
            read_fixture("test_xmlio", "test_rewrite_subject_group_04_expected.xml"),
        ),
        # example adding a new subject when none existed
        (
            fixture_file("test_xmlio", "test_rewrite_subject_group_05.xml"),
            ["New Heading"],
            "display-channel",
            True,
            read_fixture("test_xmlio", "test_rewrite_subject_group_05_expected.xml"),
        ),
        # example removes the subject tags if an empty list of subjects is provided
        (
            fixture_file("test_xmlio", "test_rewrite_subject_group_06.xml"),
            [],
            "display-channel",
            True,
            read_fixture("test_xmlio", "test_rewrite_subject_group_06_expected.xml"),
        ),
    )
    def test_rewrite_subject_group(
        self, xml, subjects, subject_group_type, overwrite, xml_expected
    ):
        root = xmlio.parse(xml)
        xmlio.rewrite_subject_group(root, subjects, subject_group_type, overwrite)
        # unicode encoding option added in python 3
        rough_string = ElementTree.tostring(root, encoding="unicode")
        self.assertEqual(rough_string, xml_expected)

    def test_reparsed_tag(self):
        """test parsing a string into xml.dom"""
        tag_name = "root"
        tag_string = (
            'Test &amp; <i class="a">a</i> <mml:math><mml:mi>K</mml:mi></mml:math>.'
        )
        reparsed = xmlio.reparsed_tag(tag_name, tag_string)
        self.assertEqual(
            reparsed.toprettyxml(indent=""),
            read_fixture("test_xmlio", "test_reparsed_tag_01_expected.xml"),
        )


@ddt
class TestParse(unittest.TestCase):
    @unpack
    @data(
        (
            (
                b'<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article PUBLIC'
                b' "-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD v1.1 20151215//EN"'
                b'  "JATS-archivearticle1.dtd">'
                b"<article/>"
            ),
            {
                "name": "article",
                "pubid": (
                    "-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD"
                    " v1.1 20151215//EN"
                ),
                "system": "JATS-archivearticle1.dtd",
            },
        )
    )
    def test_parse_doctype(self, xml, doctype_dict_expected):
        root, doctype_dict = xmlio.parse(BytesIO(xml), True)
        self.assertEqual(doctype_dict, doctype_dict_expected)

    @unpack
    @data(
        (
            (
                b'<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article PUBLIC'
                b' "-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD v1.1 20151215//EN"'
                b'  "JATS-archivearticle1.dtd">'
                b"<?covid-19-tdm?>"
                b"<article/>"
            ),
            "covid-19-tdm",
        )
    )
    def test_parse_doctype_processing(self, xml, pi_target_expected):
        root, doctype_dict, processing_instruction_nodes = xmlio.parse(
            BytesIO(xml), True, True
        )
        self.assertEqual(processing_instruction_nodes[0].target, pi_target_expected)

    @unpack
    @data(
        (
            (
                b'<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article PUBLIC'
                b' "-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD v1.1 20151215//EN"'
                b'  "JATS-archivearticle1.dtd">'
                b"<?covid-19-tdm ?>"
                b"<article>\n<p/>\n<?fig-width 50%?>\n<fig/>\n</article>"
            ),
            "covid-19-tdm",
        )
    )
    def test_parse_doctype_insert_pis(self, xml, pi_target_expected):
        "test parse XML including to insert processing instruction nodes"
        root, doctype_dict, processing_instructions = xmlio.parse(
            BytesIO(xml), True, True, insert_pis=True
        )
        xml_string = xmlio.output(root,
        output_type=None,
        doctype_dict=doctype_dict,
        processing_instructions=processing_instructions,)
        self.assertEqual(xml_string, xml)


@ddt
class TestOutput(unittest.TestCase):
    @unpack
    @data(
        (
            b"<article/>",
            "JATS",
            (
                '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article PUBLIC '
                '"-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD v1.1d3'
                ' 20150301//EN"  "JATS-archivearticle1.dtd">'
                "<article/>"
            ),
        ),
        (
            b"<article/>",
            None,
            '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article><article/>',
        ),
    )
    def test_output(self, xml, doc_type, xml_expected):
        root = xmlio.parse(BytesIO(xml))
        xml_output = xmlio.output(root, doc_type)
        self.assertEqual(xml_output.decode("utf-8"), xml_expected)

    @unpack
    @data(
        (
            (
                b'<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article PUBLIC'
                b' "-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange DTD v1.1 20151215//EN"'
                b'  "JATS-archivearticle1.dtd">'
                b"<?covid-19-tdm?>"
                b"<article/>"
            ),
            "JATS",
            (
                '<?xml version="1.0" encoding="UTF-8"?><!DOCTYPE article '
                'PUBLIC "-//NLM//DTD JATS (Z39.96) Journal Archiving and Interchange '
                'DTD v1.1 20151215//EN"  "JATS-archivearticle1.dtd">'
                "<?covid-19-tdm ?>"
                "<article/>"
            ),
        )
    )
    def test_output_processing_instructions(self, xml, doc_type, xml_expected):
        root, doctype_dict, processing_instructions = xmlio.parse(
            BytesIO(xml), True, True
        )
        xml_output = xmlio.output(root, doc_type, doctype_dict, processing_instructions)
        self.assertEqual(xml_output.decode("utf-8"), xml_expected)


if __name__ == "__main__":
    unittest.main()
